from __future__ import annotations
import os
from regendoc.actions import Action
from typing import Callable, Iterator, Sequence
import contextlib
import tempfile
from pathlib import Path
from .parse import parse_actions, correct_content
from .substitute import SubstituteAddress, SubstituteRegex, default_substituters
import logging
from rich.logging import RichHandler
from rich.console import Console
from rich.progress import BarColumn, Progress, TextColumn, TimeRemainingColumn

log = logging.getLogger("regendoc")
log.propagate = False

NORMALIZERS = Sequence[Callable[[str], str]]


def normalize_content(content: str, operators: NORMALIZERS) -> str:
    lines = content.splitlines(True)
    result = []
    for line in lines:
        for op in operators:
            line = op(line)
        result.append(line)
    return "".join(result)


def check_file(
    path: Path, content: list[str], tmp_dir: Path, normalize: NORMALIZERS
) -> list[Action]:
    needed_updates = []
    for action in parse_actions(content, file=path):

        new_content = action(tmp_dir)
        if new_content:
            action.new_content = normalize_content(new_content, normalize)
            needed_updates.append(action)
    return needed_updates


def print_diff(action: Action, console: Console) -> None:
    content, out = action.content, action.new_content
    assert out is not None

    if out != content:
        import difflib

        differ = difflib.Differ()
        outl = out.splitlines(True)
        contl = content.splitlines(True)
        lines = differ.compare(contl, outl)

        styles = {"+": "bold green", "-": "bold red", "?": "bold blue"}
        if lines:
            console.print("$", action.target, style="bold blue")
        for line in lines:
            style = styles.get(line[0])
            console.print(line, style=style, end="")


def mktemp(rootdir: Path | None, name: str) -> Path:
    if rootdir is not None:
        return Path(rootdir)
    root = tempfile.gettempdir()
    rootdir = Path(root, "regendoc-tmpdirs")
    rootdir.mkdir(exist_ok=True)
    return Path(tempfile.mkdtemp(prefix=name + "-", dir=rootdir))


@contextlib.contextmanager
def ux_setup(verbose: bool) -> Iterator[Progress]:

    columns = [
        TextColumn("[progress.description]{task.description}"),
        TextColumn("[progress.percentage]{task.completed:3.0f}/{task.total:3.0f}"),
        BarColumn(),
        TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
        TimeRemainingColumn(),
    ]
    with Progress(
        *columns,
    ) as progress:

        handler = RichHandler(
            markup=True,
            rich_tracebacks=True,
            show_path=False,
            log_time_format="[%H:%m]",
            console=progress.console,
        )
        log.addHandler(handler)
        log.setLevel(logging.DEBUG if verbose else logging.INFO)

        yield progress


def run(
    files: list[Path],
    update: bool,
    normalize: list[SubstituteRegex | SubstituteAddress] | None = None,
    rootdir: Path | None = None,
    def_name: str | None = None,
    verbose: bool = False,
) -> None:
    parsed_normalize = normalize or []

    cwd = Path.cwd()
    tmpdir: Path = mktemp(rootdir, cwd.name)

    with ux_setup(verbose) as progress:
        task_id = progress.add_task(description="progressing files")
        for num, name in enumerate(progress.track(files, task_id=task_id)):

            targetdir = tmpdir.joinpath("%s-%d" % (os.path.basename(name), num))
            with open(name) as fp:
                content = list(fp)
            targetdir.mkdir()
            log.info(f"[bold]{name}[/bold]")
            updates = check_file(
                path=name,
                content=content,
                tmp_dir=targetdir,
                normalize=default_substituters(targetdir) + parsed_normalize,
            )
            for action in updates:
                print_diff(action, progress.console)
            if update:
                corrected = correct_content(content, updates)
                with open(name, "w") as f:
                    f.writelines(corrected)



from pathlib import Path

from typing import List, Optional, Union
import typer
from . import run
from .substitute import SubstituteRegex, SubstituteAddress

app = typer.Typer()


@app.command()
def typer_main(
    files: List[Path],
    update: bool = typer.Option(False, "--update"),
    normalize: List[str] = typer.Option(default=[]),
    rootdir: Optional[Path] = None,
    def_name: Optional[str] = None,
    verbose: bool = typer.Option(False, "--verbose"),
) -> None:

    parsed_normalize: List[Union[SubstituteRegex, SubstituteAddress]] = [
        SubstituteRegex.parse(s) for s in normalize
    ]
    run(
        files=files,
        update=update,
        normalize=parsed_normalize,
        rootdir=rootdir,
        def_name=def_name,
        verbose=verbose,
    )



from __future__ import annotations

import re

from itertools import count
from collections import defaultdict
import sys
import os
from pathlib import Path


class SubstituteRegex:
    def __init__(self, match: str, replace: str) -> None:
        try:
            self.match = re.compile(match)
        except re.error as e:
            raise ValueError(match) from e
        self.replace = replace

    @classmethod
    def parse(cls: type[SubstituteRegex], s: str) -> SubstituteRegex:
        parts = s.split(s[0])
        assert len(parts) == 4
        return cls(match=parts[1], replace=parts[2])

    def __call__(self, line: str) -> str:
        return self.match.sub(self.replace, line)

    def __repr__(self) -> str:
        return "<Substituter {self.match.pattern!r} to {self.replace!r}>".format(
            self=self
        )


class SubstituteAddress:
    def __init__(self) -> None:
        self.match = re.compile(r"at 0x([0-9a-f]+)>")
        self.counter = count(1)
        self.known_addresses: dict[str, int] = defaultdict(lambda: next(self.counter))

    def replace_address(self, pattern_match: re.Match[str]) -> str:
        group = pattern_match.group(1)

        return f"at 0xdeadbeef{self.known_addresses[group]:04x}>"

    def __call__(self, line: str) -> str:
        return self.match.sub(self.replace_address, line)


def default_substituters(targetdir: Path) -> list[SubstituteAddress | SubstituteRegex]:
    return [
        SubstituteRegex(
            match=re.escape(os.fspath(targetdir)), replace="/path/to/example"
        ),
        SubstituteRegex(match=re.escape(os.getcwd()), replace="$PWD"),
        SubstituteAddress(),
        SubstituteRegex(
            match=re.escape(sys.prefix) + ".*?site-packages", replace="$PYTHON_SITE"
        ),
        SubstituteRegex(match=re.escape(sys.prefix), replace="$PYTHON_PREFIX"),
    ]



from __future__ import annotations
import os
import subprocess
import shutil
from pathlib import Path

from dataclasses import dataclass
from typing import Callable
from logging import getLogger

log = getLogger(__name__)


def write(target_dir: Path, action: Action) -> None:
    # XXX: insecure
    assert not action.target.startswith(os.path.sep), action.target
    log.debug("write to [bold]%s[/]", action.target)
    target = target_dir.joinpath(action.target)
    target.parent.mkdir(exist_ok=True, parents=True)
    target.write_text(action.content)


def process(target_dir: Path, action: Action) -> str:
    if action.cwd:

        # the cwd option is insecure and used for examples
        # that already have all files in place
        # like an examples folder for example
        if action.cwd is not None and action.cwd == Path("."):
            assert action.file is not None
            src = action.file.parent
            target_dir = target_dir / "CWD"
        else:
            src = action.file.parent.joinpath(action.cwd)

            target_dir = target_dir.joinpath(action.cwd)
        shutil.copytree(src, target_dir)

    if not os.path.isdir(target_dir):
        os.makedirs(target_dir)
    target = action.target
    log.debug("popen %r\n  cwd=%s", target, target_dir)
    output = subprocess.run(
        target,
        shell=True,
        cwd=target_dir,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=0,
        encoding="utf-8",
    )

    return output.stdout


def wipe(target_dir: Path, action: Action) -> None:
    log.debug("wiping targetdir [bold warning]%s[/]", target_dir)
    shutil.rmtree(target_dir)
    os.mkdir(target_dir)


ACTIONS = {"shell": process, "wipe": wipe, "write": write}

COMMAND_TYPE = Callable[[Path, "Action"], "str|None"]


@dataclass
class Action:
    command: COMMAND_TYPE
    target: str
    content: str
    file: Path
    new_content: str | None = None
    cwd: Path | None = None
    indent: int = 0
    line: int = 0

    def __call__(self, target_dir: Path) -> str | None:
        return self.command(target_dir, self)



from __future__ import annotations
import os
import pytest
from pathlib import Path
import subprocess
from regendoc.parse import blocks, correct_content, classify, parse_actions
from regendoc.actions import Action, process, write

from regendoc import run, check_file


@pytest.fixture
def fake_file(tmp_path: Path) -> Path:
    return tmp_path / "mess.rst"


input_data_for_blocks = """
some text
more text

next block::

    indent
      block

    with more

.. directive:: test
    :param: test

    text
"""


simple = """
an echo:
    $ echo hi
    oh no
"""

simple_corrected = simple.replace("oh no", "hi")


indented = r"""
an extra echo:
    $ printf "hi\n\nextra hi\n"
    hi

    extra hi
"""


def test_blocks() -> None:
    result = blocks(input_data_for_blocks.splitlines(True))
    expected = [
        # indent level, start line, lines
        (0, 1, ["some text\n", "more text\n", "\n", "next block::\n"]),
        (4, 6, ["indent\n", "  block\n", "\n", "with more\n"]),
        (0, 11, [".. directive:: test\n"]),
        (4, 12, [":param: test\n", "\n", "text\n"]),
    ]
    assert result == expected


def test_classify_write(fake_file: Path) -> None:

    write_result = classify(
        ["# content of test_foo.py\n", "def test()\n", "    pass\n"], file=fake_file
    )

    expected = Action(
        command=write,
        cwd=None,
        target="test_foo.py",
        content="def test()\n    pass\n",
        indent=0,
        line=0,
        file=fake_file,
    )
    assert write_result == expected


def test_classify_shell(fake_file: Path) -> None:
    cmd = classify(lines=["$ py.test -x\n", "crud\n"], file=fake_file)
    expected = Action(
        command=process,
        cwd=None,
        target="py.test -x",
        content="crud\n",
        indent=0,
        line=0,
        file=fake_file,
    )
    assert cmd == expected


def test_classify_chdir_shell(fake_file: Path) -> None:
    cmd = classify(lines=["testing $ echo hi\n", "crud\n"], file=fake_file)

    expected = Action(
        command=process,
        cwd=Path("testing"),
        target="echo hi",
        content="crud\n",
        indent=0,
        line=0,
        file=fake_file,
    )

    assert cmd == expected


EXAMPLE_DOC = """\
a simple test function call with one argument factory
==============================================================
the function argument.  Let's look at a simple self-contained
example that you can put into a test module::
    # content of: test_simplefactory.py
    def pytest_funcarg__myfuncarg(request):
        return 42
    def test_function(myfuncarg):
        assert myfuncarg == 17
.. code-block:: bash
    $ py.test test_simplefactory.py
    output should be here - but thats nice for testing
the end.
"""


@pytest.fixture
def example(tmp_path: Path) -> Path:
    p = tmp_path.joinpath("example.txt")

    p.write_text(EXAMPLE_DOC)
    return p


def test_simple_new_content(tmp_path: Path) -> None:

    (needed_update,) = check_file(
        path=Path("example"),
        content=simple.splitlines(True),
        tmp_dir=tmp_path,
        normalize=[],
    )

    expected_update = Action(
        command=process,
        cwd=None,
        file=Path("example"),
        target="echo hi",
        content="oh no\n",
        new_content="hi\n",
        indent=4,
        line=2,
    )
    print(needed_update)
    assert needed_update == expected_update


def test_single_update(fake_file: Path) -> None:
    update = Action(
        command=process,
        target="echo hi",
        content="oh no\n",
        new_content="hi\n",
        indent=4,
        line=2,
        file=fake_file,
    )

    corrected = correct_content(simple.splitlines(True), [update])
    assert corrected == simple_corrected.splitlines(True)


def test_stripped_whitespace_no_update(tmp_path: Path) -> None:
    raw = indented.splitlines(True)
    (action,) = check_file(
        path=Path("example"), content=raw, tmp_dir=tmp_path, normalize=[]
    )
    assert action.content == action.new_content
    corrected = correct_content(raw, [action])
    assert raw == corrected


def test_actions_of(example: Path, fake_file: Path) -> None:

    actions = list(parse_actions(example.read_text().splitlines(), file=fake_file))

    interesting = [(action.command, action.target) for action in actions]
    expected = [
        (write, "test_simplefactory.py"),
        (process, "py.test test_simplefactory.py"),
    ]

    assert interesting == expected


def test_check_file(tmp_path: Path, example: Path) -> None:

    check_file(
        path=Path("test.txt"),
        content=example.read_text().splitlines(True),
        tmp_dir=tmp_path,
        normalize=[],
    )
    assert tmp_path.joinpath("test_simplefactory.py").is_file()


def test_main_no_update(tmp_path: Path, example: Path) -> None:
    run([example], update=False, rootdir=tmp_path)
    # check for the created tmpdir

    assert tmp_path.joinpath("example.txt-0").is_dir()


def test_empty_update(tmp_path: Path) -> None:
    simple_fp = tmp_path / "simple.txt"
    simple_fp.write_text("")
    run([simple_fp], update=True, rootdir=tmp_path)
    corrected = simple_fp.read_text()
    assert corrected == ""


def test_main_update(tmp_path: Path) -> None:
    simple_fp = tmp_path / "simple.txt"
    simple_fp.write_text(simple)
    run([simple_fp], update=True, rootdir=tmp_path)
    corrected = simple_fp.read_text()

    assert corrected == simple_corrected


def test_docfile_chdir(tmp_path: Path, monkeypatch: pytest.MonkeyPatch) -> None:
    path = tmp_path / "nested/file"
    path.parent.mkdir()

    path.write_text("some text\n")
    monkeypatch.chdir(path.parent)
    filename = tmp_path / "test.txt"
    content = ["some shell test\n", "  nested $ cat file\n", "  some other text\n"]

    (action,) = parse_actions(content, file=filename)
    excpected_action = Action(
        command=process,
        content="some other text\n",
        cwd=Path("nested"),
        file=filename,
        indent=2,
        line=1,
        target="cat file",
    )
    assert action == excpected_action

    needed_updates = check_file(
        path=tmp_path / "example.txt",
        content=content,
        tmp_dir=tmp_path / "tmp",
        normalize=[],
    )

    assert needed_updates
    # is it copied?
    assert tmp_path.joinpath("tmp/nested/file").is_file()


def test_parsing_problem(tmp_path: Path) -> None:
    simple_fp = tmp_path.joinpath("index.txt")
    simple_fp.write_text(example_index)

    run([simple_fp], update=True, rootdir=tmp_path)
    corrected = simple_fp.read_text()
    assert corrected == example_index


example_index = """
py.test: no-boilerplate testing with Python
==============================================

.. toctree::
"""


def test_wipe(tmp_path: Path) -> None:
    fp = tmp_path / "example.txt"
    fp.write_text(".. regendoc:wipe\n")
    check_file(
        path=fp, content=fp.read_text().splitlines(), tmp_dir=tmp_path, normalize=[]
    )

    assert not list(tmp_path.iterdir())


def test_dotcwd(tmp_path: Path) -> None:
    name = tmp_path / "src/file"
    name.parent.mkdir()
    name.touch()
    tmp = tmp_path / "tmp"
    tmp.mkdir()
    check_file(
        path=name,
        content=["   . $ echo hi"],
        tmp_dir=tmp,
        normalize=[],
    )


def test_cli(tmp_path: Path) -> None:
    root = tmp_path.joinpath("tmp")
    root.mkdir()
    sample = tmp_path.joinpath("example.txt")
    sample.write_text(simple)
    subprocess.check_call(["regendoc", os.fspath(sample), "--rootdir", os.fspath(root)])



from __future__ import annotations
from typing import NamedTuple, cast
from pathlib import Path
from .actions import COMMAND_TYPE, Action, write, process, wipe
from typing import Iterable


class LineBlock(NamedTuple):
    last_indent: int
    firstline: int
    items: list[str]


def dedent(line: str, last_indent: int) -> tuple[int, str]:
    if line[:last_indent].isspace():
        return last_indent, line[last_indent:]
    stripped = line.lstrip(" ")
    return len(line) - len(stripped), stripped


def blocks(lines: list[str]) -> list[LineBlock]:
    if not lines:
        return []
    result: list[LineBlock] = []
    firstline = 0
    last_indent = dedent(lines[0], 0)[0]
    items: list[str] = []
    for lineno, line in enumerate(lines):

        indent, rest = dedent(line, last_indent)

        if lineno and indent != last_indent:
            if items[0] == "\n":
                del items[0]
                firstline += 1
            if items and items[-1] == "\n":
                del items[-1]
            result.append(LineBlock(last_indent, firstline, items))
            items = [rest]
            last_indent = indent
            firstline = lineno

        else:
            last_indent = indent
            items.append(rest or "\n")

    else:
        result.append(LineBlock(indent, firstline, items))
    return result


def correct_content(content: Iterable[str], updates: list[Action]) -> list[str]:

    lines = list(content)
    for update in reversed(updates):
        line = update.line
        old_lines = len(update.content.splitlines())
        indent = " " * update.indent
        new_lines = [
            (indent + _line).rstrip() + "\n"
            for _line in cast(str, update.new_content).splitlines(True)
        ]
        lines[line + 1 : line + old_lines + 1] = new_lines

    return lines


def classify(
    lines: list[str], file: Path, indent: int = 0, line: int = 0
) -> Action | None:
    if not lines:
        return None
    first = lines[0]
    content = "".join(lines[1:])

    def at(command: COMMAND_TYPE, target: str, cwd: Path | None = None) -> Action:
        return Action(
            command=command,
            cwd=cwd,
            target=target,
            content=content,
            indent=indent,
            line=line,
            file=file,
        )

    if first.startswith("# content of"):
        target = first.strip().split()[-1]
        return at(write, target)
    elif first[0] == "$":
        cmd = first[1:].strip()
        return at(process, cmd)
    elif " $ " in first:
        cwd, target = first.split(" $ ")
        target = target.strip()
        return at(process, target, Path(cwd))
    elif not indent and any(".. regendoc:wipe" in x for x in lines):
        return at(wipe, "")

    return None


def parse_actions(content: list[str] | str, file: Path) -> Iterable[Action]:
    lines = content if isinstance(content, list) else content.splitlines(True)
    for indent, line, data in blocks(lines):
        action = classify(lines=data, indent=indent, line=line, file=file)
        if action is not None:
            yield action
